package com.example.useencryption;

import java.io.ByteArrayInputStream;
import java.io.ByteArrayOutputStream;
import java.io.IOException;
import java.security.InvalidAlgorithmParameterException;
import java.security.InvalidKeyException;
import java.security.NoSuchAlgorithmException;
import java.security.SecureRandom;
import java.security.spec.InvalidKeySpecException;
import java.security.spec.KeySpec;

import javax.crypto.BadPaddingException;
import javax.crypto.Cipher;
import javax.crypto.IllegalBlockSizeException;
import javax.crypto.NoSuchPaddingException;
import javax.crypto.SecretKey;
import javax.crypto.SecretKeyFactory;
import javax.crypto.spec.IvParameterSpec;
import javax.crypto.spec.PBEKeySpec;
import javax.crypto.spec.SecretKeySpec;

public class CryptUtil {
    /* 鍵生成アルゴリズム */
    private static final String KEYGEN_ALGORITHM = "PBKDF2WithHmacSHA1";
    /* 鍵生成時の変換回数、多い程解析されにくくなるが処理時間が長くなる */
    private static final int ITERATION_COUNT = 1024;
    /* 鍵長、AESの場合は128、192、256 (bit) のいずれか */
    private static final int KEY_LENGTH = 256;
    /* ブロック長、AESは16バイト */
    private static final int BLOCK_LENGTH = 16;
    /* 暗号化アルゴリズム */
    private static final String ENCRYPTION_ALGORITHM = "AES/CBC/PKCS5Padding";

    /* 暗号化 */
    public static byte[] encryptData(char[] password, byte[] rawData) throws NoSuchAlgorithmException, NoSuchPaddingException,
            InvalidKeyException, InvalidAlgorithmParameterException,
            IllegalBlockSizeException, BadPaddingException, InvalidKeySpecException {
        // SALT, IV, 暗号化データを書き込んでいくStream
        ByteArrayOutputStream bout = new ByteArrayOutputStream();
        // SALTとIVの生成に使う乱数
        SecureRandom random = new SecureRandom();
        // SALTは鍵長に合わせる（単位は鍵長はbit, SALTはbyteなので8で割る）
        byte[] salt = new byte[KEY_LENGTH / 8];
        random.nextBytes(salt);
        // パスワードにSALTをかける
        KeySpec keySpec = new PBEKeySpec(password, salt, ITERATION_COUNT, KEY_LENGTH);
        // 暗号化アルゴリズムとパスワードを指定して鍵を生成する
        SecretKeyFactory factory = SecretKeyFactory.getInstance(KEYGEN_ALGORITHM);
        SecretKey secretKey = new SecretKeySpec(factory.generateSecret(keySpec).getEncoded(), "AES");
        // 暗号化の準備
        Cipher cipher = Cipher.getInstance(ENCRYPTION_ALGORITHM);
        // IVも暗号化の度に乱数を使う
        byte[] iv = new byte[BLOCK_LENGTH];
        random.nextBytes(iv);
        cipher.init(Cipher.ENCRYPT_MODE, secretKey, new IvParameterSpec(iv));
        try {
            bout.write(salt);
            bout.write(iv);
            // 以下はdoFinal()で一気に処理する方法
            // データを細切れにして処理する方法はdecryptData()を参照してください
            byte[] cipherOutput = cipher.doFinal(rawData);
            if(cipherOutput != null) {    // 基本的には発生しない
                bout.write(cipherOutput);
            }
            bout.flush();
            return bout.toByteArray();
        } catch (IOException e) {
            e.printStackTrace();
        } finally {
            if(bout != null)try{bout.close();}catch(Exception ignore){}
        }
        return null;
    }

    /* 復号 */
    public static byte[] decryptData(char[] password, byte[] encData) throws NoSuchAlgorithmException, NoSuchPaddingException,
            InvalidKeyException, InvalidAlgorithmParameterException,
            IllegalBlockSizeException, BadPaddingException, InvalidKeySpecException {
        // サイズの大きいデータが渡された場合でも、復号時に使用するメモリ量を抑えるためStreamを使用する
        ByteArrayInputStream bin = new ByteArrayInputStream(encData);
        // 復号された値を書き加えていくStream
        ByteArrayOutputStream bout = new ByteArrayOutputStream();
        try{
            // データの先頭バイトをSALTとして取り出す
            byte[] salt = new byte[KEY_LENGTH / 8];
            bin.read(salt);
            // 次の16バイトをIVとして取り出す。AESはブロック長が16バイトなため
            byte[] iv = new byte[BLOCK_LENGTH];
            bin.read(iv);
            // パスワードにSALTをかける
            KeySpec keySpec = new PBEKeySpec(password, salt, ITERATION_COUNT, KEY_LENGTH);
            // 暗号化アルゴリズムとパスワードを指定して鍵を生成する
            SecretKeyFactory factory = SecretKeyFactory.getInstance(KEYGEN_ALGORITHM);
            SecretKey secretKey = new SecretKeySpec(factory.generateSecret(keySpec).getEncoded(), "AES");
            // 復号の準備
            Cipher cipher = Cipher.getInstance(ENCRYPTION_ALGORITHM);
            cipher.init(Cipher.DECRYPT_MODE, secretKey, new IvParameterSpec(iv));
            int len = 0;
            byte[] byteBuffer = new byte[10240];
            byte[] cipherOutput = null;
            // encDataの値をStream経由で少しずつ読み込み、復号していく
            // 小さいデータならStreamを使わずdoFinal()に直接byte配列を渡す方法も可能
            // encryptData()を参考に
            while((len = bin.read(byteBuffer)) > 0){
                cipherOutput = cipher.update(byteBuffer, 0, len);
                if(cipherOutput != null){
                    bout.write(cipherOutput);
                }
            }
            // 最後のブロックにパディングして返却
            cipherOutput = cipher.doFinal();
            if(cipherOutput != null){
                bout.write(cipherOutput);
            }
            bout.flush();
            return bout.toByteArray();
        } catch (IOException e) {
            e.printStackTrace();
        } finally {
            if(bin != null)try{bin.close();}catch(Exception ignore){}
            if(bout != null)try{bout.close();}catch(Exception ignore){}
        }
        return null;
    }
}
